# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

__all_models__ = [
    "graph-network",
    "transformer",
    "equivariant-transformer",
    "tensornet",
]
